import collections
import functools
import logging
import re

import simplejson

from . import magicnumbers
from . import util
from . import filter
from .xpath import xpb


log = logging.getLogger(__name__)


looking_for_filters = filter.Filters()
looking_for_filters.register_filter_builder(filter.gentation_filter)
looking_for_filters.register_filter_builder(filter.location_filter)
looking_for_filters.register_filter_builder(
    filter.age_filter,
    decider=filter.Filters.any_not_none_decider
)
@looking_for_filters.register_filter_builder(decider=filter.Filters.any_decider)
def status_filter(status):
    return '7,{0}'.format(int(status))


class LookingFor(object):
    """Represent the looking for attributes belonging to an okcupid.com
    profile.
    """

    _ages_re = re.compile(u'\s*Ages ([0-9]{1,3})\u2013([0-9]{1,3})\s*')
    _ages_re2 = re.compile(u'\s*Age ([0-9]{1,3})')

    _looking_for_xpb = xpb.div.with_classes('text', 'what_i_want')

    def __init__(self, profile):
        self._profile = profile

    @util.cached_property
    def raw_fields(self):
        li_elements = self._looking_for_xpb.li.apply_(
            self._profile.profile_tree
        )
        return {li.attrib['id'].split('_')[1]: li.text_content()
                for li in li_elements}

    def update_property(function):
        @property
        @functools.wraps(function)
        def wrapped(self):
            return function(self)

        @wrapped.setter
        def wrapped_setter(self, value):
            self.update(**{function.__name__: value})

        return wrapped_setter

    @update_property
    def gentation(self):
        """The sex/orientation that the user is looking for."""
        return self.raw_fields.get('gentation').lower().strip()

    @update_property
    def ages(self):
        """The age range that the user is interested in."""
        match = self._ages_re.match(self.raw_fields.get('ages'))
        if not match:
            match = self._ages_re2.match(self.raw_fields.get('ages'))
            return self.Ages(int(match.group(1)),int(match.group(1)))
        return self.Ages(int(match.group(1)), int(match.group(2)))

    @update_property
    def single(self):
        """Whether or not the user is only interested in people that are single.
        """
        return 'display: none;' not in self._looking_for_xpb.li(id='ajax_single').\
            one_(self._profile.profile_tree).attrib['style']

    @update_property
    def near_me(self):
        """Whether the user is only interested in people that are close to them.
        """
        return 'near' in self.raw_fields.get('near', '').lower().strip()

    @update_property
    def kinds(self):
        """The kinds of relationship tha the user is looking for."""
        return self.raw_fields.get('lookingfor', '').\
            replace('For', '').strip().split(', ')

    def update(self, ages=None, single=None, near_me=None, kinds=None,
               gentation=None):
        """Update the looking for attributes of the logged in user.

        :param ages: The ages that the logged in user is interested in.
        :type ages: tuple
        :param single: Whether or not the user is only interested in people that
                       are single.
        :type single: bool
        :param near_me: Whether or not the user is only interested in
                        people that are near them.
        :type near_me: bool
        :param kinds: What kinds of relationships the user should be updated to
                      be interested in.
        :type kinds: list
        :param gentation: The sex/orientation of people the user is interested
                          in.
        :type gentation: str
        """
        ages = ages or self.ages
        single = single if single is not None else self.single
        near_me = near_me if near_me is not None else self.near_me
        kinds = kinds or self.kinds
        gentation = gentation or self.gentation
        data = {
            'okc_api': '1',
            'searchprefs.submit': '1',
            'update_prefs': '1',
            'lquery': '',
            'locid': '0',
            'filter5': '1, 1' # TODO(@IvanMalison) Do this better...
        }
        if kinds:
            kinds_numbers = self._build_kinds_numbers(kinds)
            if kinds_numbers:
                data['lookingfor'] = kinds_numbers
        age_min, age_max = ages
        data.update(looking_for_filters.legacy_build(
            status=single, gentation=gentation, radius=25 if near_me else 0,
            age_min=age_min, age_max=age_max
        ))
        log.info(simplejson.dumps({'looking_for_update': data}))
        util.cached_property.bust_caches(self)
        response = self._profile.authcode_post('profileedit2', data=data)
        self._profile.refresh(reload=False)
        return response.content

    @staticmethod
    def _build_kinds_numbers(looking_for_items):
        looking_for_numbers = []
        if not isinstance(looking_for_items, collections.Iterable):
            looking_for_numbers = (looking_for_items,)
        for item in looking_for_items:
            for matcher, number in magicnumbers.looking_for_re_numbers:
                if matcher.search(item) is not None:
                    looking_for_numbers.append(number)
        return looking_for_numbers

    del update_property
    Ages = collections.namedtuple('Ages', ('min', 'max'))



import time
import logging

import six

from . import helpers
from . import util
from .attractiveness_finder import AttractivenessFinder
from .json_search import SearchFetchable, search
from .location import LocationQueryCache
from .messaging import ThreadFetcher, MessageThread
from .photo import PhotoUploader
from .profile import Profile
from .profile_copy import Copy
from .question import Questions
from .session import Session
from .xpath import xpb


log = logging.getLogger(__name__)


class User(object):
    """Encapsulate a logged in okcupid user."""

    @classmethod
    def from_credentials(cls, username, password):
        """
        :param username: The username to log in with.
        :type username: str
        :param password: The password to log in with.
        :type password: str
        """
        return cls(Session.login(username, password))

    _visitors_xpb = xpb.div.with_class('user_info').\
                    div.with_class('profile_info').div.with_class('username').\
                    a.with_class('name').text_

    _visitors_current_page_xpb = xpb.div.with_class('pages').\
                                 span.with_class('curpage').text_
    _visitors_total_page_xpb = xpb.div.with_class('pages').\
                               a.with_class('last').text_

    def __init__(self, session=None):
        """
        :param session: The session which will be used for interacting
                        with okcupid.com
                        If none is provided, one will be instantiated
                        automatically with the credentials in
                        :mod:`~okcupyd.settings`
        :type session: :class:`~okcupyd.session.Session`
        """
        self._session = session or Session.login()
        self._message_sender = helpers.Messager(self._session)
        assert self._session.log_in_name is not None, (
            "The session provided to the user constructor must be logged in."
        )
        #: A :class:`~okcupyd.profile.Profile` object belonging to the logged
        #: in user.
        self.profile = Profile(self._session, self._session.log_in_name)

        #: A :class:`~okcupyd.util.fetchable.Fetchable` of
        #: :class:`~okcupyd.messaging.MessageThread` objects corresponding to
        #: messages that are currently in the user's inbox.
        self.inbox = util.Fetchable(ThreadFetcher(self._session, 1))
        #: A :class:`~okcupyd.util.fetchable.Fetchable` of
        #: :class:`~okcupyd.messaging.MessageThread` objects corresponding to
        #: messages that are currently in the user's outbox.
        self.outbox = util.Fetchable(ThreadFetcher(self._session, 2))
        #: A :class:`~okcupyd.util.fetchable.Fetchable` of
        #: :class:`~okcupyd.messaging.MessageThread` objects corresponding to
        #: messages that are currently in the user's drafts folder.
        self.drafts = util.Fetchable(ThreadFetcher(self._session, 4))

        #: A :class:`~okcupyd.util.fetchable.Fetchable` of
        #: :class:`~okcupyd.profile.Profile` objects of okcupid.com users that
        #: have visited the user's profile.
        self.visitors = util.Fetchable.fetch_marshall(
            util.GETFetcher(self._session, 'visitors',
                            lambda start_at: {'low': start_at}),
            util.PaginationProcessor(
                lambda user: Profile(self._session, user), self._visitors_xpb,
                self._visitors_current_page_xpb, self._visitors_total_page_xpb,
            )
        )

        #: A :class:`~okcupyd.question.Questions` object that is instantiated
        #: with the owning :class:`~.User` instance's session.
        self.questions = Questions(self._session)

        #: An :class:`~okcupyd.attractiveness_finder._AttractivenessFinder`
        #: object that is instantiated with the owning :class:`~.User`
        #: instance's session.
        self.attractiveness_finder = AttractivenessFinder(self._session)

        #: A :class:`~okcupyd.photo.PhotoUploader` that is instantiated with
        #: the owning :class:`~.User` instance's session.
        self.photo = PhotoUploader(self._session)

        #: A :class:`~okcupyd.location.LocationQueryCache` instance
        self.location_cache = LocationQueryCache(self._session)

    def get_profile(self, username):
        """Get the :class:`~okcupyd.profile.Profile` associated with the
        supplied username.

        :param username: The username of the profile to retrieve.
        """
        return self._session.get_profile(username)

    @property
    def username(self):
        """Return the username associated with the :class:`.User`."""
        return self.profile.username

    def message(self, username, message_text):
        """Message an okcupid user. If an existing conversation between the
        logged in user and the target user can be found, reply to that thread
        instead of starting a new one.

        :param username: The username of the user to which the message should
                         be sent.
        :type username: str
        :param message_text: The body of the message.
        :type message_text: str
        """
        # Try to reply to an existing thread.
        if not isinstance(username, six.string_types):
            username = username.username
        for mailbox in (self.inbox, self.outbox):
            for thread in mailbox:
                if thread.correspondent.lower() == username.lower():
                    thread.reply(message_text)
                    return

        return self._message_sender.send(username, message_text)

    def search(self, **kwargs):
        """Call :func:`~okcupyd.json_search.SearchFetchable` to get a
        :class:`~okcupyd.util.fetchable.Fetchable` object that will lazily
        perform okcupid searches to provide :class:`~okcupyd.profile.Profile`
        objects matching the search criteria.

        Defaults for `gender`, `gentation`, `location` and `radius` will
        be provided if none are given.

        :param kwargs: See the :func:`~okcupyd.json_search.SearchFetchable`
                       docstring for details about what parameters are
                       available.
        """
        kwargs.setdefault('gender', self.profile.gender[0])
        gentation = helpers.get_default_gentation(self.profile.gender,
                                                  self.profile.orientation)
        kwargs.setdefault('gentation', gentation)

        # We are no longer setting location by default because the
        # query api that okcupid offers seems to have gotten much
        # stricter. This should be revisited so that search gives good
        # default results. This is issue #85.

        # kwargs.setdefault('location', self.profile.location)

        kwargs.setdefault('radius', 25)
        kwargs.setdefault('location_cache', self.location_cache)
        if 'count' in kwargs:
            count = kwargs.pop('count')
            return search(session=self._session, count=count, **kwargs)
        return SearchFetchable(self._session, **kwargs)

    def delete_threads(self, thread_ids_or_threads):
        """Call :meth:`~okcupyd.messaging.MessageThread.delete_threads`.

        :param thread_ids_or_threads: A list whose members are either
                                      :class:`~.MessageThread` instances
                                      or okc_ids of message threads.
        """
        return MessageThread.delete_threads(self._session,
                                            thread_ids_or_threads)

    def get_user_question(self, question, fast=False,
                          bust_questions_cache=False):
        """Get a :class:`~okcupyd.question.UserQuestion` corresponding to the
        given :class:`~okcupyd.question.Question`.

        HUGE CAVEATS: If the logged in user has not answered the relevant
        question, it will automatically be answered with whatever the first
        answer to the question is.

        For the sake of reducing the number of requests made when
        this function is called repeatedly this function does not
        bust the cache of this :class:`~.User`'s
        :attr:`okcupyd.profile.Profile.questions` attribute. That means that
        a question that HAS been answered could still be answered by
        this function if this :class:`~.User`'s
        :attr:`~okcupyd.profile.P:attr:`okcupyd.profile.Profile.questions`
        was populated previously (This population happens automatically --
        See :class:`~okcupyd.util.fetchable.Fetchable` for details about when
        and how this happens).

        :param question: The question for which a
                         :class:`~okcupyd.question.UserQuestion` should
                         be retrieved.
        :type question: :class:`~okcupyd.question.BaseQuestion`
        :param fast: Don't try to look through the users existing questions to
                     see if arbitrarily answering the question can be avoided.
        :type fast: bool
        :param bust_questions_cache: clear the
                                     :attr:`~okcupyd.profile.Profile.questions`
                                     attribute of this users
                                     :class:`~okcupyd.profile.Profile`
                                     before looking for an existing answer.
                                     Be aware that even this does not eliminate
                                     all race conditions.
        :type bust_questions_cache: bool
        """
        if bust_questions_cache:
            self.profile.questions()
        user_question = None if fast else self.profile.find_question(
            question.id
        )
        if user_question is None:
            self.questions.respond(question.id, [1], [1], 3)
            # Give okcupid some time to update. I wish there were a better
            # way...
            for _ in range(10):
                start_time = time.time()
                user_question = self.profile.find_question(
                    question.id,
                    self.profile.question_fetchable(recent=1)
                )
                if user_question is None:
                    log.debug(
                        "Could not find question with id {0} in "
                        "questions.".format(question.id)
                    )
                    if time.time() - start_time < 1:
                        time.sleep(1)
                else:
                    break
        return user_question

    def get_question_answer_id(self, question, fast=False,
                               bust_questions_cache=False):
        """Get the index of the answer that was given to `question`

        See the documentation for :meth:`~.get_user_question` for important
        caveats about the use of this function.

        :param question: The question whose `answer_id` should be retrieved.
        :type question: :class:`~okcupyd.question.BaseQuestion`
        :param fast: Don't try to look through the users existing questions to
                     see if arbitrarily answering the question can be avoided.
        :type fast: bool
        :param bust_questions_cache: :param bust_questions_cache: clear the
                                     :attr:`~okcupyd.profile.Profile.questions`
                                     attribute of this users
                                     :class:`~okcupyd.profile.Profile`
                                     before looking for an existing answer.
                                     Be aware that even this does not eliminate
                                     all race conditions.
        :type bust_questions_cache: bool
        """
        if hasattr(question, 'answer_id'):
            # Guard to handle incoming user_question.
            return question.answer_id

        user_question = self.get_user_question(
            question, fast=fast, bust_questions_cache=bust_questions_cache
        )
        # Look at recently answered questions
        return user_question.get_answer_id_for_question(question)

    def quickmatch(self):
        """Return a :class:`~okcupyd.profile.Profile` obtained by visiting the
        quickmatch page.
        """
        response = self._session.okc_get('quickmatch', params={'okc_api': 1})
        return Profile(self._session, response.json()['sn'])

    def copy(self, profile_or_user):
        """Create a :class:`~okcupyd.profile_copy.Copy` instance with the
        provided object as the source and this :class:`~okcupyd.user.User`
        as the destination.

        :param profile_or_user: A :class:`~okcupyd.user.User` or
                                :class:`~okcupyd.profile.Profile` object.
        """
        return Copy(profile_or_user, self)

    def __repr__(self):
        return u'User("{0}")'.format(self.profile.username)



import getpass

import logging
import shutil
import sys

from coloredlogs import ColoredFormatter, HostNameFilter

from okcupyd import settings


DOMAIN = 'www.okcupid.com'
headers = {
    'accept': 'text/javascript, text/html, application/xml, '
    'text/xml, */*',
    'accept-encoding': 'gzip,deflate',
    'accept-language': 'en-US,en;q=0.8',
    'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
    'origin': 'https://www.okcupid.com',
    'x-prototype-version': 1.7,
    'x-requested-with': 'XMLHttpRequest',
}


def enable_logger(log_name, level=logging.DEBUG):
    log = logging.getLogger(log_name)
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(
        ColoredFormatter(level_styles={'WARNING': dict(color='red')})
    )
    log.addHandler(handler)
    HostNameFilter.install(handler=handler)

    handler.setLevel(level)
    log.setLevel(level)


def get_credentials():
    if not settings.USERNAME:
        input_function = input if sys.version_info.major == 3 else raw_input
        settings.USERNAME = input_function('username: ').strip()
    if not settings.PASSWORD:
        settings.PASSWORD = getpass.getpass('password: ')


def add_command_line_options(add_argument, use_short_options=True):
    """
    :param add_argument: The add_argument method of an ArgParser.
    :param use_short_options: Whether or not to add short options.
    """
    logger_args = ("--enable-logger",)
    credentials_args = ("--credentials",)
    if use_short_options:
        logger_args += ('-l',)
        credentials_args += ('-c',)
    add_argument(*logger_args, dest='enabled_loggers',
                 action="append", default=[],
                 help="Enable the specified logger.")
    add_argument(*credentials_args, dest='credential_files',
                 action="append", default=[],
                 help="Use the specified credentials module to update "
                 "the values in okcupyd.settings.")
    add_argument('--echo', dest='echo', action='store_true', default=False,
                 help="Echo SQL.")


def handle_command_line_options(args):
    """
    :param args: The args returned from an ArgParser
    """
    for enabled_log in args.enabled_loggers:
        enable_logger(enabled_log)
    for credential_file in args.credential_files:
        settings.load_credentials_from_filepath(credential_file)
    if args.echo:
        from okcupyd import db
        db.echo = True
        db.Session.kw['bind'].echo = True
    return args


def save_file(filename, data):
    with open(filename, 'wb') as out_file:
        shutil.copyfileobj(data, out_file)


def find_all(a_str, sub):
    start = 0
    while True:
        start = a_str.find(sub, start)
        if start == -1: return
        yield start
        start += len(sub) # use start += 1 to find overlapping matches


def replace_all_case_insensitive(a_str, sub, replacement):
    segments = []
    last_stop = 0
    for start in find_all(a_str.lower(), sub.lower()):
        segments.append(a_str[last_stop:start])
        segments.append(replacement)
        last_stop = start + len(sub)
    segments.append(a_str[last_stop:])
    return ''.join(segments)
